/*
 * Copyright (c) KATO Takenori, 1999.
 * 
 * All rights reserved.  Unpublished rights reserved under the copyright
 * laws of Japan.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 	$Id: operator.c,v 1.13 1999/07/13 11:53:49 kato Exp $
 */

#include <Xm/Xm.h>

#include <math.h>
#include <stdio.h>

#include "const.h"
#include "stack.h"
#include "keyboard.h"
#include "option.h"
#include "operator.h"

static double
ANGLE2RAD(double a)
{
	switch (angleunit) {
	case ANGLE_DEG:
		return a / 180.0 * PI;
	case ANGLE_GRAD:
		return a / 200.0 * PI;
	}
	return a;
}

static double
RAD2ANGLE(double r)
{
	switch (angleunit) {
	case ANGLE_DEG:
		return r * 180.0 / PI;
	case ANGLE_GRAD:
		return r * 200.0 / PI;
	}
	return r;
}

/*
 * Binary Operators
 */
int
op_add_double(void)
{
	double	X, Y;

	X = *((double*)stackPop());
	Y = *((double*)stackPop());
	Y += X;
	stackPush(&Y);

	return 0;
}

int
op_add_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y += X;
	stackPush(&Y);

	return 0;
}

int
op_sub_double(void)
{
	double	X, Y;

	X = *((double*)stackPop());
	Y = *((double*)stackPop());
	Y -= X;
	stackPush(&Y);

	return 0;
}

int
op_sub_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y -= X;
	stackPush(&Y);

	return 0;
}

int
op_mul_double(void)
{
	double	X, Y;

	X = *((double*)stackPop());
	Y = *((double*)stackPop());
	Y *= X;
	stackPush(&Y);

	return 0;
}

int
op_mul_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y *= X;
	stackPush(&Y);

	return 0;
}

int
op_div_double(void)
{
	double	X, Y;

	X = *((double*)stackPop());
	Y = *((double*)stackPop());
	Y /= X;
	stackPush(&Y);

	return 0;
}

int
op_div_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y /= X;
	stackPush(&Y);

	return 0;
}

int
op_pow_double(void)
{
	double	X, Y;

	X = *((double*)stackPop());
	Y = *((double*)stackPop());
	Y = pow(Y, X);
	stackPush(&Y);

	return 0;
}

int
op_pow_int(void)
{
	double	X, Y;
	int	R;

	X = (double)*((int*)stackPop());
	Y = (double)*((int*)stackPop());
	R = (int)pow(Y, X);
	stackPush(&R);

	return 0;
}

int
op_shiftR_double(void)
{
	int	X, Y;
	double	R;

	X = (int)(*((double*)stackPop()));
	Y = (int)(*((double*)stackPop()));
	R = (double)(Y >> X);
	stackPush(&R);

	return 0;
}

int
op_shiftR_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y >> X;
	stackPush(&Y);

	return 0;
}

int
op_shiftL_double(void)
{
	int	X, Y;
	double	R;

	X = (int)(*((double*)stackPop()));
	Y = (int)(*((double*)stackPop()));
	R = (double)(Y << X);
	stackPush(&R);

	return 0;
}

int
op_shiftL_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y << X;
	stackPush(&Y);

	return 0;
}

int
op_or_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y | X;
	stackPush(&Y);

	return 0;
}

int
op_and_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y & X;
	stackPush(&Y);

	return 0;
}

int
op_xor_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y ^ X;
	stackPush(&Y);

	return 0;
}

int
op_xnor_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = ~(Y ^ X);
	stackPush(&Y);

	return 0;
}

int
op_mod_int(void)
{
	int	X, Y;

	X = *((int*)stackPop());
	Y = *((int*)stackPop());
	Y = Y % X;
	stackPush(&Y);

	return 0;
}

/*
 * Unary Operators
 */
int
op_sin(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = sin(ANGLE2RAD(X));
	stackSetX(&X);

	return 0;
}

int
op_cos(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = cos(ANGLE2RAD(X));
	stackSetX(&X);

	return 0;
}

int
op_tan(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = tan(ANGLE2RAD(X));
	stackSetX(&X);

	return 0;
}

int
op_asin(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = RAD2ANGLE(asin(X));
	stackSetX(&X);

	return 0;
}

int
op_acos(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = RAD2ANGLE(acos(X));
	stackSetX(&X);

	return 0;
}

int
op_atan(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = RAD2ANGLE(atan(X));
	stackSetX(&X);

	return 0;
}

int
op_sinh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = sinh(X);
	stackSetX(&X);

	return 0;
}

int
op_cosh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = cosh(X);
	stackSetX(&X);

	return 0;
}

int
op_tanh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = tanh(X);
	stackSetX(&X);

	return 0;
}

int
op_asinh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = asinh(X);
	stackSetX(&X);

	return 0;
}

int
op_acosh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = acosh(X);
	stackSetX(&X);

	return 0;
}

int
op_atanh(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = atanh(X);
	stackSetX(&X);

	return 0;
}

int
op_log(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = log10(X);
	stackSetX(&X);

	return 0;
}

int
op_log_int(void)
{
	double	X;
	int	R;

	X = (double)*((int*)stackGetX());
	R = (int)log10(X);
	stackSetX(&R);

	return 0;
}

int
op_antilog_double(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = pow(10.0, X);
	stackSetX(&X);

	return 0;
}

int
op_antilog_int(void)
{
	double	X;
	int	R;

	X = (double)*((int*)stackGetX());
	R = (int)(pow(10.0, X));
	stackSetX(&R);

	return 0;
}

int
op_ln(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = log(X);
	stackSetX(&X);

	return 0;
}

int
op_antiln(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = exp(X);
	stackSetX(&X);

	return 0;
}

int
op_sqrt_double(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = sqrt(X);
	stackSetX(&X);

	return 0;
}

int
op_sqrt_int(void)
{
	double	X;
	int	R;

	X = (double)*((int*)stackGetX());
	R = (int)sqrt(X);
	stackSetX(&R);

	return 0;
}

int
op_cbrt_double(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = cbrt(X);
	stackSetX(&X);

	return 0;
}

int
op_cbrt_int(void)
{
	double	X;
	int	R;

	X = (double)*((int*)stackGetX());
	R = (int)cbrt(X);
	stackSetX(&R);

	return 0;
}

int
op_square_double(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = X * X;
	stackSetX(&X);

	return 0;
}

int
op_square_int(void)
{
	int	X;

	X = *((int*)stackGetX());
	X = X * X;
	stackSetX(&X);

	return 0;
}

int
op_recip(void)
{
	double	X;

	X = *((double*)stackGetX());
	X = 1.0 / X;
	stackSetX(&X);

	return 0;
}

int
op_facto_double(void)
{
#ifdef	FACTO_INTEGER
	unsigned long	X, i, res;
	double	d;

	X = *((double*)stackGetX());
	for (i = 1, res = 1; i<= X; i++)
		res *= i;
	d = (double)res;
	stackSetX(&d);
#else
	double	X, a, b, tmp, s;
	static double	c[6] = {
	    76.18009172947146, -86.50532032941677,
	    24.01409824083091, -1.23173957245015,
	    0.1208650973866179e-2, -0.5395239384953e-5
	};
	int	i;

	/* Approximation by Lanczos */
	a = b = X = *((double*)stackGetX()) + 1.0;
	tmp = a + 5.5;
	tmp -= (a + 0.5) * log(tmp);
	s = 1.000000000190015;
	for (i = 0; i < 6; i++)
		s += c[i] / ++b;
	X = -tmp + log(2.5066282746310005 * s / a);
	X = exp(X);
	stackSetX(&X);
#endif
	return 0;
}

int
op_facto_int(void)
{
#ifdef	FACTO_INTEGER
	unsigned int	X, i, res;

	X = *((unsigned int*)stackGetX());
	for (i = 1, res = 1; i <= X; i++)
		res *= i;

	stackSetX(&res);
#else
	int	R;
	double	X, a, b, tmp, s;
	static double	c[6] = {
	    76.18009172947146, -86.50532032941677,
	    24.01409824083091, -1.23173957245015,
	    0.1208650973866179e-2, -0.5395239384953e-5
	};
	int	i;

	/* Approximation by Lanczos */
	a = b = X = (double)(*((int*)stackGetX())) + 1.0;
	tmp = a + 5.5;
	tmp -= (a + 0.5) * log(tmp);
	s = 1.000000000190015;
	for (i = 0; i < 6; i++)
		s += c[i] / ++b;
	X = -tmp + log(2.5066282746310005 * s / a);
	X = exp(X);

	R = (int)(X + .5);
	stackSetX(&R);
#endif
	return 0;
}

int
op_inc_int(void)
{
	int	X;

	X = *((int*)stackGetX());
	X++;
	stackSetX(&X);

	return 0;
}

int
op_dec_int(void)
{
	int	X;

	X = *((int*)stackGetX());
	X--;
	stackSetX(&X);

	return 0;
}

int
op_neg2_int(void)
{
	int	X;

	X = *((int*)stackGetX());
	X = ~X;
	X++;
	stackSetX(&X);

	return 0;
}

int
op_not_int(void)
{
	int	X;

	X = *((int*)stackGetX());
	X = ~X;
	stackSetX(&X);

	return 0;
}
